'' =================================================================================================
''
''   File....... jm_menus_demo.spin
''   Purpose.... 
''   Author..... Jon "JonnyMac" McPhalen 
''   E-mail..... 
''   Started.... 
''   Updated.... 19 NOV 2016
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
                                                                 
  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

                                                      
con { io pins }                                                   
                                                                  
  RX1      = 31  { i }                                          ' programming / terminal
  TX1      = 30  { o }                                           
                                                                 
  SDA      = 29  { io }                                         ' eeprom / i2c
  SCL      = 28  { io }                                          

  LCD_D7   = 24  { io }                                         ' not used as button
  LCD_D6   = 23  { io }                                         ' forward button   
  LCD_D5   = 22  { io }                                         ' reverse button    
  LCD_D4   = 21  { io }                                         ' set button  
  LCD_E    = 20  { o }                                           
  LCD_RS   = 19  { o }                                           
  LCD_BL   = 16  { o }                                          ' backlight control                                           


con { app io }
                                                              
  BTN_F    = LCD_D6                                             ' shared with LCD                                             
  BTN_R    = LCD_D5                                              
  BTN_S    = LCD_D4


con { app constants }

  #0, PS_RUN, PS_EDIT

  M_FWD = %100                                                  ' button masks
  M_REV = %010
  M_SET = %001

  BTN_MS  = 25                                                  ' buttons debounce
  BTN_CLR = 50                                                  ' button clear timing
  BTN_SPD = 200                                                 ' delay for button held

  NO_START = 96                                                 ' start time resolution is 15 mins
  NO_RUN   =  0



  
  BKL_MS = 10_000                                               ' keep backlight on for 10s

  BKL_FLAG = 1 << 0                                             ' mask for backlight flag


obj

' main                                                          ' *    master Spin cog
  time       : "jm_time_80"                                     '      for timing/delays (80MHz)
  tbacklight : "jm_time_80"                                     '      for timing backlight
  io         : "jm_io"                                          '      essential io contorl
  ee         : "jm_24xx512"                                     '      eeprom access
  lcd        : "jm_lcd4"                                        '      parallel LCD output
' background

' * uses cog when loaded

                                                    
var

  long  pgmstate
  long  pgmflags


dat
'                        0000000011111111                       ' LCD lines
  s_Splash      byte    "JonnyMacControls", 0
  s_Firmware    byte    "Sprinkle1.00    ", 0


dat { sprinkler timer data }

  Water_Days    byte    %0000000                   
  Start_Times   byte    NO_START, NO_START, NO_START, NO_START
  Station_Run   byte    NO_RUN,   NO_RUN,   NO_RUN,   NO_RUN                 
      
                                                                
pub main

  setup                                                         ' setup base objects

  update_soft_rtc                                               ' sync soft rtc to hardware rtc
  
  program_id                                                    ' show splash and firmware rev

  pgmstate := PS_RUN                                            ' power up in run mode
  
  repeat
    case pgmstate
      PS_RUN  : run_irrigation
      PS_EDIT : edit_settings


pub setup

'' Configure objects and IO

  time.start                                                    ' setup for timing & delays
                                                                 
  io.start(0, 0)                                                ' clear pins in cog

  ee.start(%000)                                                ' connect to eeprom
  
  lcd.startx(LCD_BL, LCD_E, LCD_RS, LCD_D4)                     ' start lcd
  time.pause(3)                                                      
  lcd.clear

  bcog := cognew(background, @bstack) + 1                       ' start background cog *


pub update_soft_rtc

'' Read hardware RTC
'' -- set software RTC with tod values

  { TBD -- based on selected RTC }


pub set_soft_rtc(d, h, m, s)

'' Set sofware RTC

  tix := 0                                                      ' reset seconds timer      
  rtc := (d << 24) | (h << 16) | (m << 8) | s                   ' set rtc registers


pub program_id

'' Display program splash screen and firmware id

  display_2x8(@s_splash)
  backlight_on(BKL_MS)
  time.pause(1500)

  display_2x8(@s_firmware) 
  time.pause(1000)
  clear_display(true)


pub run_irrigation | btns, last, now                            ' shows day and time only! (demo)

  clear_display(true)
  backlight_on(BKL_MS)

  last := -1

  repeat
    now := rtc                                                  ' capture rtc registers
    if (now.byte[SCS] <> last)
      lcd.home
      lcd.str(@@DayName[now.byte[DAY]])
      lcd.line(1)
      lcd_dec2(now.byte[HRS])
      lcd.out(":")
      lcd_dec2(now.byte[MNS])  
      lcd.out(":")             
      lcd_dec2(now.byte[SCS])  
      last := now.byte[SCS]            

'     if (nnow.byte[SCS])                                       ' if new second
'       check_start_time                                        '  check for active start time
  
    btns := scan_buttons(100)
    case btns
      M_FWD : backlight_on(BKL_MS)
      M_REV : backlight_on(BKL_MS)
      M_SET : pgmstate := PS_EDIT

    if (pgmstate <> PS_RUN)
      clear_buttons(BTN_CLR)
      return
      
    check_backlight


dat
'                        0000000011111111                       ' LCD lines
  s_MainMenu    byte    "SET     W-Days  ", 0
                byte    "SET     W-Starts", 0
                byte    "SET     Sta Run ", 0
                byte    "SET     Clock   ", 0    
                byte    "EXIT    Setup   ", 0       
  
    
pub edit_settings | idx

  backlight_on(posx)                                            ' backlight on "forever"   

  idx := 0                                                      ' start with watering days

  repeat
    if (pgmstate <> PS_EDIT)                                    ' return to run mode when done
      return
   
    idx := select_from_list2(idx, 0, 4, @s_MainMenu)            ' select section to edit
    case idx                     
      0 : set_watering_days
      1 : set_irrigation_starts   
      2 : set_station_run_times 
      3 : set_clock
      4 : pgmstate := PS_RUN


pub set_watering_days | d, check

'' Set watering day (0..6) status to No (0) or Yes (1)

  repeat d from 0 to 6
    clear_display(true)                                         ' update display
    lcd.str(@s_Run)
    lcd.str(@@DayName[d])
     
    check := (Water_Days >> d) & 1                              ' get current setting
    check := select_from_list1(check, 0, 1, 1, @s_No, 4)        ' allow No/Yes edit
    if (check == 1)                                               ' update program var
      Water_Days |= 1 << d
    else
      Water_Days &= !(1 << d)

  ee.wr_byte(@Water_Days, Water_Days)                           ' save to eeprom
    
  
pub set_irrigation_starts | n, idx, stime

'' Set the start times

  repeat n from 0 to 3
    clear_display(true)                                         ' update display
    lcd.str(@s_Start)                                            
    lcd_dec1(n+1)                                               ' display as 1..4
                                                                 
    idx := Start_Times[n]                                       ' get index
                                                                 
    repeat                                                       
      lcd.line(1)                                                
      if (idx => 96)                                            ' max?
        lcd.str(@s_Off)                                         ' Off + clean-up
        lcd.out(" ")                                             
        lcd.out(" ")
      else
        stime := idx * 15                                       ' convert to time  
        lcd_dec2(stime / 60)
        lcd.out(":")
        lcd_dec2(stime // 60)
     
      if (get_input(@idx, 0, 96))                               ' 0..96 = 00:00 .. 24:00 (off)
        clear_buttons(BTN_CLR)  
        quit
      else
        time.pause(BTN_SPD)

    Start_Times[n] := idx                   
    ee.wr_byte(@Start_Times[n], idx)
     
     
pub set_station_run_times | n, rtmins

'' Set station run times

  repeat n from 0 to 3
    clear_display(true)                                         ' update display
    lcd.str(@s_Station)
    lcd_dec1(n+1)                                               ' display as 1..4
     
    rtmins := Station_Run[n]                                    ' get current time
    
    repeat
      lcd.line(1)                                               ' display time
      if (rtmins == NO_RUN)                                     ' if 0
        lcd.str(@s_Off)                                         '  display "OFF" 
        lcd.out(" ")
      else                                                      ' else
        lcd_dec1(rtmins / 60)                                   '  display as H:MM
        lcd.out(":")
        lcd_dec2(rtmins // 60)
        
      if (get_input(@rtmins, 0, 240))                           ' edit, 0:00 .. 4:00
        clear_buttons(BTN_CLR) 
        quit
      else
        time.pause(BTN_SPD) 
      
    Station_Run[n] := rtmins                                    ' save program variable
    ee.wr_byte(@Station_Run[n], rtmins)                         ' write to eeprom
    
    
pub set_clock  

  rtc := set_day(rtc)  
  rtc := set_hours(rtc)
  rtc := set_minutes(rtc)
  rtc := set_seconds(rtc) 


pub set_day(now)

  display_screen(true, @s_Today, false)

  now.byte[3] := select_from_list1(now.byte[3], 0, 6, 1, @s_Sun, 4)

  return now       
  
  
pub set_hours(now)

'' Set/Edit hours field of now

  display_screen(true, @s_Hours, false)
  lcd.line(1)
  show_clock(now)

  now.byte[2] := edit_value(now.byte[2], 0, 23, 0, 1, 2)

  return now


pub set_minutes(now)

'' Set/Edit minutes field of now

  display_screen(true, @s_Minutes, false)
  lcd.line(1)
  show_clock(now)

  now.byte[1] := edit_value(now.byte[1], 0, 59, 3, 1, 2)

  return now


pub set_seconds(now)

'' Set/Edit minutes field of now

  display_screen(true, @s_Seconds, false)
  lcd.line(1)
  show_clock(now)

  now.byte[0] := edit_value(now.byte[0], 0, 59, 6, 1, 2)

  return now


pub show_clock(tregs)

  lcd_dec2(tregs.byte[2])
  lcd.out(":")
  lcd_dec2(tregs.byte[1])
  lcd.out(":")
  lcd_dec2(tregs.byte[0])


con

  { --------------- }
  {  E D I T O R S  }
  { --------------- }


pub select_from_list1(idx, lo, hi, ln, p_list, slen) | last

'' For selecting item from 0-indexed list
'' -- idx is the starting value
'' -- hi is maximum index value
'' -- ln is the lcd line (0 or 1) to display list
'' -- p_list is pointer to list
'' -- slen is length of strings (+ terminator)

  idx := lo #> idx <# hi                                        ' fix in range

  repeat
    last := -1                                                  ' force lcd refresh
    repeat
      if (idx <> last)                                          ' if new idx, update lcd
        lcd.line(ln)
        lcd.substr(p_list + (idx * slen), slen)
        last := idx

      if (get_input(@idx, lo, hi))                              ' get a button
        clear_buttons(BTN_CLR)                                  ' clear if Set/Enter pressed
        return idx                                              ' retunr current index
      else
        time.pause(BTN_SPD)                                     ' control speed for held button          


pub select_from_list2(idx, lo, hi, p_list) | last

'' For selecting 2-line screen from list
'' -- lo and hi define index limits
'' -- p_list is pointer to list strings (16-chars for both lines)

  idx := lo #> idx <# hi                                  

  repeat
    last := -1                                           
    repeat
      if (idx <> last)                    
        display_2x8(p_list + (idx * 17))  
        last := idx                       

      if (get_input(@idx, lo, hi))        
        clear_buttons(BTN_CLR)            
        return idx                        
      else
        time.pause(BTN_SPD)               


pub edit_value(value, lo, hi, col, line, width) | last

'' For selecting 1- to 3-digit value in range lo..hi
'' -- col and line control placement in lcd
'' -- width controls characters printed

  value := lo #> value <# hi                                    ' fix in range
                                                                 
  last := !value                                                ' force lcd refresh   
                                                                 
  repeat                                                        ' loop through values
    if (value <> last)                                          ' if value changed
      lcd.crsr_xy(col, line)                                    ' update display
      lcd_dec(value, width)
      last := value
   
    if (get_input(@value, lo, hi) > 0)                          ' get input, mod value
      clear_buttons(BTN_CLR)
      return value                          
    else
      time.pause(BTN_SPD)


pub get_input(p_value, lo, hi) | btns

  repeat
    btns := scan_buttons(BTN_MS)                                ' check buttons
  until (btns <> %000)                                          ' repeat until press

  case btns
    M_FWD :
      if (++long[p_value] > hi)  
        long[p_value] := lo      
      return 0                   

    M_REV :
      if (--long[p_value] < lo) 
        long[p_value] := hi     
      return 0                  

    M_SET :
      return 1


con

  { ----------------------------- }
  {  B U T T O N   H E L P E R S  }
  { ----------------------------- }


pub scan_buttons(ms) | btns, t

'' Check buttons for input, debounce ms milliseconds

  dira[BTN_F..BTN_S] := %000                                    ' make inputs  
  btns := ina[BTN_F..BTN_S]

  if (btns)                                                     ' if something pressed
    t := cnt                                                    ' sync debounce timing
    repeat ms                                                   ' debounce
      waitcnt(t += MS_001)                                      ' hold 1ms
      btns &= ina[BTN_F..BTN_S]                                 ' re-scan
    return btns

      
pub clear_buttons(ms) | timer, t

'' Holds program until all buttons are released for ms milliseconds

  dira[BTN_F..BTN_S] := %000                                    ' make inputs
  
  timer := 0                                                    ' reset timer

  t := cnt
  repeat
    waitcnt(t += MS_001)                                        ' wait 1ms           
    if (ina[BTN_F..BTN_S] == %000)                              ' are buttons clear
      if (++timer == ms)                                        ' increment and check limit
        return
    else
      timer := 0                                                ' no clear, restart timer

      
con

  { ------------------------------- }
  {  D I S P L A Y   H E L P E R S  }
  { ------------------------------- }


dat { common strings }

  s_No          byte    "No ", 0
  s_Yes         byte    "Yes", 0

  s_Off         byte    "Off", 0
  s_On          byte    "On ", 0

  s_Midnight    byte    "00:00:00", 0
  s_Hours       byte    "Hours", 0
  s_Minutes     byte    "Mins", 0
  s_Seconds     byte    "Secs", 0
  s_Today       byte    "Today is", 0

  s_Run         byte    "Run ", 0
  s_Start       byte    "Start ", 0
  s_Station     byte    "Sta ", 0

  s_Sun         byte    "Sun", 0
  s_Mon         byte    "Mon", 0
  s_Tue         byte    "Tue", 0
  s_Wed         byte    "Wed", 0
  s_Thu         byte    "Thu", 0
  s_Fri         byte    "Fri", 0
  s_Sat         byte    "Sat", 0

  DayName       Word    @s_Sun, @s_Mon, @s_Tue, @s_Wed, @s_Thu, @s_Fri, @s_Sat              
  

pub clear_display(write)

'' Clear LCD
'' -- if write is true, allow LCD to finish CLS

  lcd.clear
  if (write)                                                    ' do we want to write on return?
    time.pause(3)                                               ' yes, let LCD finish


pub display_2x8(p_str)

'' Fill 2x8 with 16-character string at p_str
'' -- short string may cause undesired result

  lcd.home
  lcd.substr(p_str+0, 8)
  lcd.line(1)
  lcd.substr(p_str+8, 8)
  

pub display_screen(clr, p_str0, p_str1)

'' Put pstr_0 or pstr_1 into display
'' -- if clr is true, display is cleared
''    * use if strings less than 8 chars

  if (clr)
    clear_display(true)

  if (p_str0 > 0)
    line_out(0, p_str0) 

  if (p_str1 > 0)
    line_out(1, p_str1) 

  
pub line_out(line, p_str)

'' Writes string at p_str to specified line of display
'' -- valid only for lines 0 and 1

  lcd.line(line)    
  lcd.str(p_str)


pub list_str(p_list, n)

'' Writes nth string from list at p_list

  repeat n
    p_list += strsize(p_list) + 1

  lcd.str(p_list)


pub lcd_dec(value, width)

  if (width == 1)
    lcd_dec1(value)
  elseif (width == 2)
    lcd_dec2(value)
  elseif (width == 3)
    lcd_dec3(value)


pub lcd_dec1(value)

'' Write 1-digit decimal value
'' -- must be positive

  lcd.out("0" + value)
  

pub lcd_dec2(value)

'' Write 2-digit value with leading 0

  lcd_dec1(value  / 10)
  lcd_dec1(value // 10)


pub lcd_dec3(value)

'' Write 3-digit value with leading 0s

  lcd_dec1(value  / 100)
  lcd_dec2(value // 100)


con

  { ----------------------------------- }
  {  B A C K L I G H T   C O N T R O L  }
  { ----------------------------------- }


pub check_backlight | blm

'' Check backlight timer for expiration
'' -- expired timer kills backlight

  blm := tbacklight.millis

  if (blm < 0)                                                  ' still running?
    return
  elseif (pgmflags & BKL_FLAG)                                  ' flag set?
    if (blm => 0)                                               '  and timer expired
      backlight_off


pub backlight_on(ms)

  lcd.backlight(1)                                              ' turn backlight on
  tbacklight.set(-ms)                                           ' set auto-shutdown timer
  pgmflags |= BKL_FLAG                                          ' set flag


pub backlight_off

  lcd.backlight(0)                                              ' turn backlight off
  tbacklight.set(0)                                             ' clear auto-shutdown timer
  pgmflags &= !BKL_FLAG                                         ' clear flag


con

  { ----------------------------- }
  {  B A C K G R O U N D   C O G  }
  {  - global milliseconds        }
  {  - soft rtc                   }
  { ----------------------------- }
  

var

  long  bcog                                                    ' cog running background code
  long  bstack[64]                                              ' stack space for background cog

  long  millis                                                  ' running milliseconds

                                                             
pri background | t                                              ' launch with cognew()                                           
                                                                 
  millis := 0                                                    
                                                                 
  t := cnt                                                      ' sync loop timer
  repeat                                                         
    waitcnt(t += MS_001)                                        ' run loop every millisecond
    ++millis                                                    ' advance timer                                                   
    update_rtc


con

  #0, SCS, MNS, HRS, DAY
  

var

  long  tix                                                     ' seconds timing
  long  rtc                                                     ' rtc registers
                                                                 
                                             
pri update_rtc

  if (++tix == 1000)                                            ' one second elapsed?
    tix := 0                                                    ' yes, update rtc registers
    if (++rtc.byte[SCS] == 60)                                  
      rtc.byte[SCS] := 0
      if (++rtc.byte[MNS] == 60) 
        rtc.byte[MNS] := 0
        if (++rtc.byte[HRS] == 24) 
          rtc.byte[HRS] := 0       
          if (++rtc.byte[DAY] == 7)    
            rtc.byte[DAY] := 0       


dat { license }                                                  

{{

  Copyright (C) 2016 Jon McPhalen

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}